##
# File:  ServiceWorkerBase.py
# Date:  7-July-2016
#
# Updates:
#    22-Sep-2016  jdw add session tracking
#    26-Sep-2016  jdw add _getServiceTrackingSummary()
#     2-Dec-2016  jdw include remote_addr in all tracking records.
#    18-Feb-2017  jdw use internal method to obtain siteId.
#    15-Mar-2017  jdw add trackHistory=True to _getSession()
##
"""
Base class for supporting web service processing modules.

This software was developed as part of the World Wide Protein Data Bank
Common Deposition and Annotation System Project

Copyright (c) wwPDB

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

# import os
# import sys
import time

# import types
# import string
# import traceback
# import ntpath

import logging

#
from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.ws_utils.ServiceDataStore import ServiceDataStore

from wwpdb.utils.ws_utils.ServiceSessionState import ServiceSessionState
from wwpdb.utils.ws_utils.ServiceHistory import ServiceHistory

#
logger = logging.getLogger()


class ServiceWorkerBase(object):
    def __init__(self, reqObj=None, sessionDataPrefix=None):
        """
        Base class supporting web application worker methods.

        Performs URL -> application mapping for this module.

        """
        self._reqObj = reqObj
        self._sObj = None
        self._sessionId = None
        self._sessionPath = None
        self._rltvSessionPath = None
        #
        #
        self._siteId = self._reqObj.getSiteId()
        self._cI = ConfigInfo(self._siteId)
        #
        #  ServiceDataStore prefix for general session data -- used by _getSession()
        self._sdsPrefix = sessionDataPrefix if sessionDataPrefix else "general"
        self._sds = None
        #
        # Service items include:
        # self.__class__.__name__,sys._getframe().f_code.co_name
        self.__appPathD = {}
        #

    def _trackServiceStatus(self, op, **params):
        """Add a service history status tracking record.

        :param string  op:  operation status  (e.g., 'created', 'submitted', 'completed', 'failed' , ...)
        :param dictionary params:  optional key-value payload stored with status record.

        """
        if params and "remote_addr" not in params:
            params["remote_addr"] = self._reqObj.getValue("remote_addr")
        #
        sH = ServiceHistory(historyPath=self._reqObj.getSessionUserPath())
        return sH.add(sessionId=self._sessionId, statusOp=op, **params)

    def _getServiceActivitySummary(self):
        """Get the service activity summary."""
        sH = ServiceHistory(historyPath=self._reqObj.getSessionUserPath())
        return sH.getActivitySummary()

    def addService(self, url, opName):
        self.__appPathD[url] = opName

    def addServices(self, serviceDict):
        for k, v in serviceDict.items():
            self.__appPathD[k] = v

    def _run(self, reqPath=None):
        """Map operation to path and invoke operation.  Exceptions are caught within this method.

        :returns:

        Operation output is packaged in a ServiceSessionState() object.

        """
        #
        requestPath = None
        try:
            requestPath = reqPath if reqPath else self._reqObj.getRequestPath()
            #
            if requestPath not in self.__appPathD:
                # bail out if operation is unknown -
                sst = ServiceSessionState()
                sst.setServiceError(msg="Unknown operation")
            else:
                mth = getattr(self, self.__appPathD[requestPath], None)
                sst = mth()
            return sst
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("FAILING for requestPath %r ", requestPath)
            sst = ServiceSessionState()
            sst.setServiceError(msg="Operation failure")

        return sst

    #
    def _appendSessionStore(self, iD=None):
        """Dictionary of key value pairs will be appended to the session parameter store.

        list and dict type values are extended/appended to existing values of corresponding types.


        """
        try:
            if iD is not None and isinstance(iD, dict) and len(iD) > 0:
                self._sds.update(iD)
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("FAILED updating with input %r ", iD)
        return False

    def _getSessionStoreDict(self):
        """Recover session store data as a dictionary."""
        try:
            return self._sds.getDictionary()
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("FAILED to recover session store")

        return {}

    def _setSessionStoreValue(self, ky, val):
        """Set session store data as a dictionary."""
        try:
            return self._sds.set(ky, val)
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("FAILED to set session store value")

        return {}

    def _trackSessionHistory(self, msg="ok"):
        rP = self._reqObj.getRequestPath()
        tS = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        self._sds.append("session_history", (rP, tS, msg))

    def _getSession(self, new=False, useContext=False, contextOverWrite=True, trackHistory=True):
        """Join existing session or create new session as required."""
        #
        try:
            #
            self._sObj = self._reqObj.getSessionObj(new=new)
            self._sessionId = self._sObj.getId()
            self._sessionPath = self._sObj.getPath()
            logging.debug("session   id  %s", self._sessionId)
            logging.debug("session path  %s", self._sessionPath)
            if self._sessionPath is None:
                return False
            self._rltvSessionPath = self._sObj.getRelativePath()
            self._sds = ServiceDataStore(sessionPath=self._sessionPath, prefix=self._sdsPrefix)
            if useContext and not new:
                dd = self._getSessionStoreDict()
                logging.debug("Imported %r", dd)
                self._reqObj.setDictionary(dd, overWrite=contextOverWrite)
            if trackHistory:
                self._trackSessionHistory(msg="begins")
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("FAILING create or joining session")
        return False

    ##



##
# File:  ServiceSmtpUtils.py
# Date:  3-Aug-2016
#
# Updates:
#   3-Aug-2016 jdw  -- refactor from TokenUtils
##
"""
SMTP/Mail utilities

This software was developed as part of the World Wide Protein Data Bank
Common Deposition and Annotation System Project

Copyright (c) wwPDB

This software is provided under a Creative Commons Attribution 3.0 Unported
License described at http://creativecommons.org/licenses/by/3.0/.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import os
import os.path


import smtplib
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart

from email import encoders
from email.utils import formatdate


import logging

logger = logging.getLogger()


class ServiceSmtpUtils(object):
    def __init__(self):
        """Collection of mail handling methods -"""

    def sendFile(self, srcPath, toAddr, fromAddr, subject, replyAddr="noreply@mail.wwpdb.org"):
        """Internal method to mail file as text.

        Reply-To: noreply@example.com
        """

        # Create a text/plain message
        fp = open(srcPath, "rb")
        msg = MIMEText(fp.read())
        fp.close()
        #
        msg["Subject"] = subject
        msg["From"] = fromAddr
        msg["To"] = toAddr
        if replyAddr is not None:
            msg["reply-to"] = replyAddr

        #
        s = smtplib.SMTP()
        s.connect()
        s.sendmail(fromAddr, [toAddr], msg.as_string())
        s.close()

    def emailFiles(self, fromAddr, toAddr, subject, text, replyAddr=None, fileList=None, textAsAttachment=None, textAttachmentName="token.txt"):
        """'noreply@mail.wwpdb.org'"""

        msg = MIMEMultipart()
        msg["From"] = fromAddr
        msg["To"] = toAddr
        msg["Date"] = formatdate(localtime=True)
        msg["Subject"] = subject
        if replyAddr:
            msg["reply-to"] = replyAddr
        #
        msg.attach(MIMEText(text))

        for file in fileList or []:
            part = MIMEBase("application", "octet-stream")
            with open(file, "rb") as infile:
                part.set_payload(infile.read())
            encoders.encode_base64(part)
            part.add_header("Content-Disposition", 'attachment; filename="%s"' % os.path.basename(file))
            msg.attach(part)
        #
        if textAsAttachment:
            part = MIMEBase("application", "octet-stream")
            part.set_payload(textAsAttachment)
            encoders.encode_base64(part)
            part.add_header("Content-Disposition", 'attachment; filename="%s"' % textAttachmentName)
            msg.attach(part)
        #
        try:
            s = smtplib.SMTP()
            s.set_debuglevel(1)
            s.connect()
            s.sendmail(fromAddr, [toAddr], msg.as_string())
            s.close()
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("FAILING")
        return False

    def emailTextWithAttachment(self, fromAddr, toAddr, replyAddr, subject, text, textAsAttachment=None, textFileName=None):
        """"""

        msg = MIMEMultipart()
        msg["From"] = fromAddr
        msg["To"] = toAddr
        msg["Date"] = formatdate(localtime=True)
        msg["Subject"] = subject
        msg["reply-to"] = replyAddr
        #
        msg.attach(MIMEText(text))

        if textAsAttachment:
            part = MIMEBase("application", "octet-stream")
            part.set_payload(textAsAttachment)
            encoders.encode_base64(part)
            part.add_header("Content-Disposition", 'attachment; filename="%s"' % os.path.basename(textFileName))
            msg.attach(part)
        #
        try:
            s = smtplib.SMTP()
            s.set_debuglevel(1)
            s.connect()
            s.sendmail(fromAddr, [toAddr], msg.as_string())
            s.close()
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("FAILING")
        return False



##
# File:    ServiceRequest.py
# Date:    2-Aug-2016  J. Westbrook
#          Fork of WebRequest adapted for service naming conventions and functions.
# Updated:
#       3-Aug-2016 jdw standardize case for wwpdb_site_id
#      23-Sep-2016 jdw add getSessionUserPath()
##
"""
WebRequest provides containers and accessors for managing request parameter information.

This is an application neutral version shared by UI modules --

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"


import sys
import os
import logging

try:
    import json
except ImportError:
    import simplejson as json

from wwpdb.utils.ws_utils.ServiceSessionFactory import ServiceSessionFactory

logger = logging.getLogger()


class ServiceRequestBase(object):

    """Base container and accessors for input and output parameters and control information."""

    def __init__(self, paramDict=None):
        #
        #  Input and storage model is dictionary of lists (e.g. dict[myKey] = [,,,])
        #  Single values are stored in the leading element of the list (e.g. dict[myKey][0])
        #
        if paramDict is None:
            paramDict = {}
        self.__dict = paramDict

    def __outputList(self):
        sL = []
        sL.append(" ++Service request contents:\n")
        for k, vL in self.__dict.items():
            sL.append("     - Key: %-35s  value(s): %r" % (k, vL))
        return sL

    def __str__(self):
        try:
            return "\n  ".join(self.__outputList())
        except:  # noqa: E722 pylint: disable=bare-except
            return ""

    def __repr__(self):
        return self.__str__()

    def printIt(self, ofh=sys.stdout):
        try:
            ofh.write("%s" % self.__str__())
        except:  # noqa: E722 pylint: disable=bare-except
            pass

    def dump(self, format="text"):  # pylint: disable=unused-argument,redefined-builtin
        try:
            return "\n   ".join(self.__outputList())
        except:  # noqa: E722 pylint: disable=bare-except
            pass

    def getJSON(self):
        return json.dumps(self.__dict)

    def setJSON(self, JSONString):
        self.__dict = json.loads(JSONString)

    def getValue(self, myKey):
        return self._getStringValue(myKey)

    def getValueOrDefault(self, myKey, default=""):
        if not self.exists(myKey):
            return default
        v = self._getStringValue(myKey)
        if len(v) < 1:
            return default
        return v

    def getValueList(self, myKey):
        return self._getStringList(myKey)

    def getRawValue(self, myKey):
        return self._getRawValue(myKey)

    def getDictionary(self):
        return self.__dict

    #
    def setValue(self, myKey, aValue):
        try:
            self.__dict[myKey] = [aValue]
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    def setValueList(self, myKey, valueList):
        try:
            self.__dict[myKey] = valueList
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    def setDictionary(self, myDict, overWrite=False):
        for k, v in myDict.items():
            if overWrite or (not self.exists(k)):
                self.setValue(k, v)
        return True

    def exists(self, myKey):
        try:
            return myKey in self.__dict
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    #
    def _getRawValue(self, myKey):
        try:
            return self.__dict[myKey][0]
        except:  # noqa: E722 pylint: disable=bare-except
            return None

    def _getStringValue(self, myKey):
        try:
            return str(self.__dict[myKey][0]).strip()
        except:  # noqa: E722 pylint: disable=bare-except
            return ""

    def _getIntegerValue(self, myKey):
        try:
            return int(self.__dict[myKey][0])
        except:  # noqa: E722 pylint: disable=bare-except
            return None

    def _getDoubleValue(self, myKey):
        try:
            return float(self.__dict[myKey][0])
        except:  # noqa: E722 pylint: disable=bare-except
            return None

    def _getStringList(self, myKey):
        try:
            return self.__dict[myKey]
        except:  # noqa: E722 pylint: disable=bare-except
            return []


class ServiceRequest(ServiceRequestBase):
    def __init__(self, paramDict):
        super(ServiceRequest, self).__init__(paramDict)
        self.__returnFormatDefault = ""
        self.__requestPrefix = ""

    def setDefaultReturnFormat(self, return_format="html"):
        self.__returnFormatDefault = return_format
        if not self.exists("return_format"):
            self.setValue("return_format", self.__returnFormatDefault)

    def setRequestPathPrefix(self, prefix):
        """Set optional request path prefix to be removed before applying application routing."""
        self.__requestPrefix = prefix

    def getRequestPathPrefix(self):
        """Get request path prefix."""
        return self.__requestPrefix

    def getRequestPath(self):
        try:
            iRp = self._getStringValue("request_path")
            if len(self.__requestPrefix) > 0:
                if iRp.startswith(self.__requestPrefix):
                    rp = iRp[len(self.__requestPrefix) :]
                    return rp
                else:
                    return iRp
            else:
                return iRp
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("Request path processing FAILED")

        return None

    def getReturnFormat(self):
        if not self.exists("return_format"):
            self.setValue("return_format", self.__returnFormatDefault)
        return self._getStringValue("return_format")

    def setReturnFormat(self, return_format="json"):
        return self.setValue("return_format", return_format)

    def getSessionId(self):
        return self._getStringValue("session_id")

    def getTopSessionPath(self):
        return self._getStringValue("top_session_path")

    def getServiceUserId(self):
        return self._getStringValue("service_user_id")

    def setServiceUserId(self, serviceUserId):
        try:
            return self.setValue("service_user_id", serviceUserId)
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    def getSessionUserPath(self):
        return self._getStringValue("session_user_path")

    def setTopSessionPath(self, pth):
        try:
            return self.setValue("top_session_path", pth)
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    #

    def setSiteId(self, siteId):
        try:
            self.setValue("wwpdb_site_id", siteId)
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            pass
        return False

    def getSiteId(self):
        return self._getStringValue("wwpdb_site_id")

    def getSessionPath(self):
        return os.path.join(self._getStringValue("top_session_path"), "sessions")

    def getSemaphore(self):
        return self._getStringValue("semaphore")

    def getSessionObj(self, new=False):
        """Get or create new session -"""
        try:
            logger.debug("Starting")
            sObj = ServiceSessionFactory()
            if self.exists("top_session_path"):
                sObj.setTopSessionPath(topSessionPath=self._getStringValue("top_session_path"))
            if self.exists("service_user_id"):
                sObj.setServiceUserId(serviceUserId=self._getStringValue("service_user_id"))
            if new:
                sObj.assignId()
                sObj.makeSessionPath()
                self.setValue("session_id", sObj.getId())
                logger.debug("Creating new session %s ", sObj.getId())
            else:
                if self.exists("session_id"):
                    logger.debug("Aquiring existing session %s ", self._getStringValue("session_id"))
                    sObj.setId(uid=self._getStringValue("session_id"))
            #
            self.setValue("session_user_path", sObj.getSessionUserPath())
            logger.debug("Completed")
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("Session acquisition/creation FAILING")

        return sObj



##
# File:    ServiceDataStore.py
# Date:    6-July-2012
#
# Updates:
#          2-Aug-2016  jdw remove dependence on request object
#                          refactor to make content 'append' only.
#                          Add locking
#         23-Sep-2016  jdw adjust logging
#         15-Mar-2017  jdw increase lock timeout -
#         15-Mar-2017  jdw gut the concurrency handling wrap internal io methods
##
"""
Provide a storage interface for miscellaneous key,value data.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

try:
    import cPickle as pickle
except ImportError:
    import pickle
import os.path
import logging

from oslo_concurrency import lockutils

logger = logging.getLogger()


class ServiceDataStore(object):
    """Provide a storage interface for miscellaneous key,value data."""

    def __init__(self, sessionPath, prefix=None):
        self.__filePrefix = prefix if prefix is not None else "general"
        self.__sessionPath = sessionPath
        self.__filePath = None
        #
        lockutils.set_defaults(self.__sessionPath)
        #
        self.__setup()

    def __setup(self):
        #
        # self.__pickleProtocol = pickle.HIGHEST_PROTOCOL
        self.__pickleProtocol = 0
        try:
            self.__filePath = os.path.join(self.__sessionPath, self.__filePrefix + "-session-store.pic")
            logger.debug("Service data store path %r", self.__filePath)
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("FAILING for filePath %r", self.__filePath)

    def __serialize(self, iD):

        try:
            with open(self.__filePath, "wb") as fb:
                pickle.dump(iD, fb, self.__pickleProtocol)
            if "status" in iD:
                logger.debug("Session %s - wrote status value %r", self.__sessionPath, iD["status"])
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Serialization failure with file %s", self.__filePath)
        return False

    def __deserialize(self):
        rD = {}
        try:
            if not os.access(self.__filePath, os.R_OK):
                logger.warning("No data store in path %r ", self.__filePath)
                return rD
        except:  # noqa: E722 pylint: disable=bare-except
            pass
        try:
            with open(self.__filePath, "rb") as fb:
                rD = pickle.load(fb)
        except Exception as e:
            logger.exception("Deserialization failure with file %s - %r", self.__filePath, str(e))

        if "status" in rD:
            logger.debug("Session %s - read dictionary %r", self.__sessionPath, rD["status"])
        return rD

    def __str__(self):
        try:
            return "\n  ".join(self.__outputList())
        except:  # noqa: E722 pylint: disable=bare-except
            return ""

    def __repr__(self):
        return self.__str__()

    def dump(self, format="text"):  # pylint: disable=redefined-builtin,unused-argument
        try:
            return "\n   ".join(self.__outputList())
        except:  # noqa: E722 pylint: disable=bare-except
            return ""

    def getFilePath(self):
        return self.__filePath

    #
    #  Getters()  reread before any access -
    #
    @lockutils.synchronized("sessiondatastore.lock", external=True)
    def __outputList(self):
        rD = self.__deserialize()
        sL = []
        sL.append("Session data store contents:")
        for k in sorted(rD.keys()):
            v = rD[k]
            sL.append("     - Key: %-35s  value(s): %r" % (k, v))
        return sL

    @lockutils.synchronized("sessiondatastore.lock", external=True)
    def get(self, key):
        try:
            rD = self.__deserialize()
            return rD[key]
        except:  # noqa: E722 pylint: disable=bare-except
            return ""

    @lockutils.synchronized("sessiondatastore.lock", external=True)
    def getDictionary(self):
        rD = self.__deserialize()
        return rD

    #
    #  Setters ()
    #
    @lockutils.synchronized("sessiondatastore.lock", external=True)
    def set(self, key, value, overWrite=True):
        try:
            rD = self.__deserialize()
            if overWrite:
                rD[key] = value
                return self.__serialize(rD)
            else:
                if key not in rD:
                    rD[key] = value
                    return self.__serialize(rD)
                else:
                    # no overwrite
                    return False
        except Exception as e:
            logger.exception("Failure of set for key %r value %r error %r", key, value, str(e))
            return False

    @lockutils.synchronized("sessiondatastore.lock", external=True)
    def update(self, uDict):
        """Update (without overwrite) objects in the first level dictionary store."""
        try:
            rD = self.__deserialize()
            for k, v in uDict.items():
                if k not in rD:
                    rD[k] = v
                else:
                    # append cases - for only values of the
                    if isinstance(rD[k], list) and isinstance(v, list):
                        rD[k].extend(v)
                    elif isinstance(rD[k], list) and not isinstance(v, list):
                        rD[k].append(v)
                    elif isinstance(rD[k], dict) and isinstance(v, dict):
                        # only add new objects to a dict type.
                        for tk, tv in v:
                            if tk not in rD[k]:
                                rD[k][tk] = tv
                    else:
                        pass

            return self.__serialize(rD)
        except Exception as e:
            logger.exception("Failure for uDict %r %r", uDict, str(e))
            return False

    @lockutils.synchronized("sessiondatastore.lock", external=True)
    def updateAll(self, uDict):
        """Update with overwrite values first level dictionary store."""
        try:
            rD = self.__deserialize()
            for k, v in uDict.items():
                rD[k] = v
            if "status" in rD:
                logger.debug("Updating status value %r", rD["status"])
            return self.__serialize(rD)
        except Exception as e:
            logger.exception("Failure for uDict %r %r", uDict, str(e))
            return False

    @lockutils.synchronized("sessiondatastore.lock", external=True)
    def append(self, key, value):
        try:
            rD = self.__deserialize()
            if key not in rD:
                rD[key] = []
            rD[key].append(value)
            return self.__serialize(rD)
        except Exception as e:
            logger.exception("Failure for key %r value %r %r", key, value, str(e))
            return False

    @lockutils.synchronized("sessiondatastore.lock", external=True)
    def extend(self, key, valueList):
        try:
            rD = self.__deserialize()
            if key not in rD:
                rD[key] = []
            rD[key].extend(valueList)
            return self.__serialize(rD)
        except Exception as e:
            logger.exception("Failure for key %r value %r %r", key, valueList, str(e))
            return False



##
# File: ServiceUtilsMisc.py
# Date:  7-Aug-2016  John Westbrook
#
# Update:
#
##
"""
Utility functions supporting web service classes.

"""

import hashlib

import logging

logger = logging.getLogger(__name__)


def getMD5(path, block_size=4096, hr=True):
    """
    Chunked MD5 function -

    Block size directly depends on the block size of your filesystem
    to avoid performances issues

    Linux Ext4 block size
        sudo /sbin/blockdev --getbsz /dev/sda1
        > Block size:               4096

    """
    md5 = hashlib.md5()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(block_size), b""):
            md5.update(chunk)
    if hr:
        return md5.hexdigest()
    return md5.digest()



##
# File:  SessionHistory
# Date:  23-Sept-2016
#
# Updated:
#        25-Sep-2016  jdw add activity summary method -
##
"""
Methods to manage service session history tracking  --

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import logging

try:
    import cPickle as pickle
except ImportError:
    import pickle
import os.path
import copy
import time
import datetime
import dateutil.parser

from operator import itemgetter
from wwpdb.utils.ws_utils.ServiceLockFile import ServiceLockFile

logger = logging.getLogger()


class ServiceHistory(object):
    """Methods to manage service session history tracking"""

    #

    def __init__(self, historyPath, useUTC=False):
        """"""
        #
        self.__useUtc = useUTC
        self.__historyPath = historyPath
        self.__filePath = None
        self.__timeOutSeconds = 2.0
        self.__retrySeconds = 0.1
        self.__unlocked_maxretry = 5  # Unlocked status history retrieval return for pickle erro
        self.__unlocked_retrySeconds = 3  # Unlocked status history return sleep
        self.__setup()

    def __setup(self):
        #
        # self.__pickleProtocol = pickle.HIGHEST_PROTOCOL
        self.__pickleProtocol = 0
        try:
            self.__filePath = os.path.join(self.__historyPath, "history-session-store.pic")
            logger.debug("Service history data store path %r", self.__filePath)
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("FAILING for filePath %r", self.__filePath)

    #
    def __serialize(self, iD, mode="wb"):
        """Internal method to write session history data to persistent store."""
        with ServiceLockFile(self.__filePath, timeoutSeconds=self.__timeOutSeconds, retrySeconds=self.__retrySeconds) as lock:  # noqa: F841 pylint: disable=unused-variable
            try:
                with open(self.__filePath, mode) as fb:
                    pickle.dump(iD, fb, self.__pickleProtocol)
                return True
            except:  # noqa: E722 pylint: disable=bare-except
                logger.exception("Serialization failure with file %s", self.__filePath)
        return False

    def __deserialize(self):
        """Internal method to recover session history data from persistent store. Locks file"""
        rD = {}
        with ServiceLockFile(self.__filePath, timeoutSeconds=self.__timeOutSeconds, retrySeconds=self.__retrySeconds) as lock:  # noqa: F841 pylint: disable=unused-variable
            return self.__deserialize_data()
        return rD

    def __deserialize_data(self, raiseExc=False):
        """Derserialize the history data in the file. Might be locked or not.
        If raiseExc set raise exception on parsing pickle error"""
        rD = {}
        try:
            if not os.access(self.__filePath, os.R_OK):
                logging.warning("No data store in path %r ", self.__filePath)
                return rD
        except:  # noqa: E722 pylint: disable=bare-except
            pass
        try:
            with open(self.__filePath, "rb") as fb:
                while True:
                    # process each record and quit at eof
                    try:
                        d = pickle.load(fb)
                        # logger.info("Read activity record %r" % d)
                        if d["sid"] not in rD:
                            rD[d["sid"]] = {}
                        rD[d["sid"]][d["op"]] = d["data"]
                    except EOFError:
                        break
        except Exception as exc:  # noqa: E722 pylint: disable=bare-except
            if raiseExc:
                logger.error("Deserialization failure with file %s", self.__filePath)
                raise exc
            logger.exception("Deserialization failure with file %s", self.__filePath)

        return rD

    def add(self, sessionId, statusOp, **params):
        """Record a service session tracking record  -

         :param sessionId:  target session identifier
         :param statusOp:   current operation for the tracking record
         :param params:     additional payload of key-values for the tracking record.

        :rtype bool: True for success or False otherwise

        """
        dd = {}
        if params:
            dd = copy.deepcopy(params)
        if self.__useUtc:
            dd["tiso"] = datetime.datetime.utcnow().isoformat()
        else:
            dd["tiso"] = datetime.datetime.now().isoformat()
        tD = {"sid": sessionId, "op": statusOp, "data": dd}
        #
        return self.__serialize(tD, mode="a+b")

    def getHistory(self, lock=True):
        """Return a dictionary image of all session tracking data for the current service user.
        If lock is True, use lockging"""
        if lock:
            return self.__deserialize()
        else:
            # Unlocked - retry on pickle parsing failure
            cnt = 0
            while (cnt < self.__unlocked_maxretry):
                try:
                    rd = self.__deserialize_data(raiseExc=True)
                    return rd
                except pickle.UnpicklingError:
                    cnt += 1
                    logging.error("Could not unpickle unlocked, retry %s", cnt)
                    time.sleep(self.__unlocked_retrySeconds)
            logging.error("Could not parse file.  return empty")
            rD = {}
            return rD

    def getActivitySummary(self):
        """Create a summary of session activity for the service user.

        :rtype dictionary:   dictionary of summary details -
        """
        tD = self.__deserialize()
        #
        rD = {}
        sessionCount = 0
        submittedCount = 0
        completedCount = 0
        failedCount = 0
        sL = []
        #
        try:
            tStart = datetime.datetime(1969, 1, 1).isoformat()  # should always have a "created" in loop
            for sId in tD:
                sD = tD[sId]
                deltaSeconds = 0
                if "created" in sD:
                    sessionCount += 1
                    tStart = sD["created"]["tiso"]
                    st = "created"
                if "submitted" in sD:
                    submittedCount += 1
                    tBegin = sD["submitted"]["tiso"]
                    st = "submitted"
                    #
                    if "failed" in sD:
                        failedCount += 1
                        tEnd = sD["failed"]["tiso"]
                        st = "failed"
                        dt = dateutil.parser.parse(tEnd) - dateutil.parser.parse(tBegin)
                        deltaSeconds = dt.total_seconds()
                    if "completed" in sD:
                        completedCount += 1
                        tEnd = sD["completed"]["tiso"]
                        st = "completed"
                        dt = dateutil.parser.parse(tEnd) - dateutil.parser.parse(tBegin)
                        deltaSeconds = dt.total_seconds()

                    sL.append((sId, tStart, st, deltaSeconds))

            ssL = sorted(sL, key=itemgetter(1))
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("summary construction failing")
        #
        rD["session_count"] = sessionCount
        rD["submitted_count"] = submittedCount
        rD["failed_count"] = failedCount
        rD["completed_count"] = completedCount
        rD["session_list"] = ssL
        return rD



##
# File:  TokenUtils.py
# Date:  10-July-2016
#
# Updates:
#   2-Aug-2016 jdw standardized error diagnostics --
#  25-Sep-2016 jdw revise exception messages
#  13-Feb-2017 jdw add token prefix to foken data store file name -
##
"""
Base class for supporting application token management.

This software was developed as part of the World Wide Protein Data Bank
Common Deposition and Annotation System Project

Copyright (c) wwPDB

This software is provided under a Creative Commons Attribution 3.0 Unported
License described at http://creativecommons.org/licenses/by/3.0/.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import os
import datetime

try:
    import cPickle as pickle
except ImportError:
    import pickle

import jwt
from oslo_concurrency import lockutils

import logging

#
from wwpdb.utils.config.ConfigInfo import ConfigInfo

logger = logging.getLogger()


class TokenUtilsBase(object):
    def __init__(self, siteId=None, tokenPrefix=None, **kwargs):
        """
        Base class supporting application token management in persistent store.
        kwargs allows for overriding ConfigInfo for testing

        """
        #
        self._cI = ConfigInfo(siteId)
        #
        if tokenPrefix is not None:
            fn = tokenPrefix + "_TOKEN_STORE.pic"
        else:
            fn = "ANONYMOUSWS_TOKEN_STORE.pic"
        #
        self.__filePath = kwargs.get("site_service_registration_dir_path")
        if not self.__filePath:
            self.__filePath = os.path.join(self._cI.get("SITE_SERVICE_REGISTRATION_DIR_PATH"), fn)
        #
        logger.debug("Assigning token store file path %r", self.__filePath)
        self.__lockDirPath = kwargs.get("site_service_registration_lockdir_path")
        if not self.__lockDirPath:
            self.__lockDirPath = self._cI.get("SITE_SERVICE_REGISTRATION_LOCKDIR_PATH", ".")
        #
        self.__tokenD = {}
        self.__emailD = {}
        self.__tokenPrefix = tokenPrefix if tokenPrefix else "WS"
        self.__pickleProtocol = 0
        #
        lockutils.set_defaults(self.__lockDirPath)
        #
        self.deserialize()

    def __makeTokenId(self, iVal):
        return "%s_%010d" % (self.__tokenPrefix, iVal)

    def __parseTokenId(self, tokenId):
        try:
            pL = tokenId.split("_")
            prefix = pL[0]
            iVal = int(pL[1])
            return (prefix, iVal)
        except:  # noqa: E722 pylint: disable=bare-except
            return (None, None)

    def getFilePath(self):
        return self.__filePath

    @lockutils.synchronized("tokenutils.serialize-lock", external=True)
    def serialize(self):
        try:
            with open(self.__filePath, "wb") as outfile:
                pickle.dump(self.__tokenD, outfile, self.__pickleProtocol)
                pickle.dump(self.__emailD, outfile, self.__pickleProtocol)
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("FAILING")
        return False

    def deserialize(self):
        try:
            with open(self.__filePath, "rb") as outfile:
                self.__tokenD = pickle.load(outfile)
                self.__emailD = pickle.load(outfile)
            logger.debug("Recovered %4d token Ids %4d e-mails", len(self.__tokenD), len(self.__emailD))
            return True
        except Exception as e:
            logger.debug("Unable to deserialize persistent token store %r - %r", self.__filePath, str(e))
        return False

    def tokenIdExists(self, tokenId):
        try:
            return tokenId in self.__tokenD
        except:  # noqa: E722 pylint: disable=bare-except
            pass
        return False

    def tokenIdEmailExists(self, email):
        try:
            return email in self.__emailD
        except:  # noqa: E722 pylint: disable=bare-except
            pass
        return False

    @lockutils.synchronized("tokenutils.transaction-lock", external=True)
    def remove(self, tokenId):
        self.serialize()
        try:
            email = self.getTokenIdEmail(tokenId)
            if email in self.__emailD:
                del self.__emailD[email]
        except:  # noqa: E722 pylint: disable=bare-except
            self.deserialize()
            return False
        try:
            del self.__tokenD[tokenId]
        except:  # noqa: E722 pylint: disable=bare-except
            self.deserialize()
            return False
        self.serialize()
        return True

    def getTokenIdEmail(self, tokenId):
        try:
            return self.__tokenD[tokenId]["email"]
        except:  # noqa: E722 pylint: disable=bare-except
            return ""

    def saveTokenId(self, tokenId, email, **kw):
        """"""
        if tokenId is None or len(tokenId) < 12 or email is None or len(email) < 3:
            return False
        try:

            tD = {"email": email}
            tD.update(kw)
            self.__tokenD[tokenId] = tD
            self.__emailD[email] = tokenId
            self.serialize()
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Token save FAILING")
        return False

    def fetchTokenId(self, email):
        """Return any existing token assigned to the input e-mail or create the next token
        identifer with this e-mail assignment.
        """
        try:
            return self.__emailD[email]
        except:  # noqa: E722 pylint: disable=bare-except
            pass
        #
        #  Create a new tokenId and save -
        #
        iVal = self.__nextTokenId()
        tokenId = self.__makeTokenId(iVal)
        self.saveTokenId(tokenId, email)

        return tokenId

    def __nextTokenId(self):
        """Return the next token identifier string (e.g. N+1 for <PREFIX>_0000000N)

        return 1-N for success or -1 for failure
        """
        try:
            if len(self.__tokenD.keys()) > 0:
                tokenMax = max(self.__tokenD.keys())
                _pr, iV = self.__parseTokenId(tokenMax)
                iV += 1
            else:
                iV = 1
            return iV
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("FAILING")

        return -1


class JwtTokenUtils(TokenUtilsBase):
    def __init__(self, siteId=None, tokenPrefix=None, **kwargs):
        """
        Token utilities for registration webservice

        """
        super(JwtTokenUtils, self).__init__(siteId=siteId, tokenPrefix=tokenPrefix, **kwargs)
        #
        # self.__inputToken = self._reqObj.getValue('authorization').split()[1]
        #
        serviceKey = self._cI.get("SITE_SERVICE_REGISTRATION_KEY", None)
        self.__serviceKey = serviceKey if serviceKey else "secretvalue"
        self.__tokenErrorCode = 401

    def parseAuth(self, authHeader):
        rD = {"errorCode": None, "errorMessage": None, "token": None, "errorFlag": False}
        parts = authHeader.split()
        if parts[0].lower() != "bearer":
            rD = {"errorCode": self.__tokenErrorCode, "errorMessage": "Authorization header must start with Bearer", "errorFlag": True}
        elif len(parts) == 1:
            rD = {"errorCode": self.__tokenErrorCode, "errorMessage": "API access token not found", "errorFlag": True}
        elif len(parts) > 2:
            rD = {"errorCode": self.__tokenErrorCode, "errorMesage": "Authorization header must be Bearer token", "errorFlag": True}
        if not rD["errorFlag"]:
            rD["token"] = parts[1]
        return rD

    def parseToken(self, token):
        """Return data payload for the input token along with any error diagnostics -"""
        return self.__getTokenData(token)

    def getToken(self, email, expireDays=30):
        """Return a tokenId and a jwt token for the input e-mail address.
        Existing tokenId's are reused with the existing email address.

        """
        tokenId = self.fetchTokenId(email)
        jwtToken = self.__create_token(tokenId, self.__serviceKey, expireDays=expireDays)
        return tokenId, jwtToken

    def __create_token(self, tokenId, secretKey, expireDays=30, algorithm="HS256", encoding="utf-8"):
        payload = {
            #        using the standard JWT claim keys --
            # subject id
            "sub": tokenId,
            # creation datetime
            "iat": datetime.datetime.utcnow(),
            # expiration datetime
            "exp": datetime.datetime.utcnow() + datetime.timedelta(days=expireDays),
        }
        token = jwt.encode(payload, secretKey, algorithm=algorithm)
        #
        # PyJwt 2.x return string
        #
        if int(jwt.__version__.split(".")[0]) >= 2:
            return token
        else:
            #
            # 'utf-8' or 'unicode_escape'?
            #
            return token.decode(encoding)

    def __parseToken(self, token, secretKey="secretvalue", algorithm="HS256"):
        return jwt.decode(token, secretKey, algorithms=algorithm)

    def __getTokenData(self, token):
        """Return token payload as a dictionary and processing diagnostics -"""
        errorCode = None
        errorMessage = None
        tokenData = {}
        #
        if token is None:
            errorMessage = "Missing token"
        try:
            tokenData = self.__parseToken(token, secretKey=self.__serviceKey)
        except jwt.DecodeError:
            errorMessage = "API access token is invalid"
            errorCode = self.__tokenErrorCode
        except jwt.ExpiredSignatureError:
            errorMessage = "API access token has expired"
            errorCode = self.__tokenErrorCode
        except:  # noqa: E722 pylint: disable=bare-except
            errorMessage = "API access token processing error"
            errorCode = self.__tokenErrorCode

        tokenData["errorCode"] = errorCode
        tokenData["errorMessage"] = errorMessage
        tokenData["errorFlag"] = errorMessage is not None

        return tokenData


class JwtTokenReader(object):
    def __init__(self, siteId=None):
        """
        Limited set of token methods required to read and validate a JWT tokens.

        """
        self._cI = ConfigInfo(siteId)
        serviceKey = self._cI.get("SITE_SERVICE_REGISTRATION_KEY", default=None)
        self.__serviceKey = serviceKey if serviceKey else "secretvalue"
        self.__tokenErrorCode = 401

    def parseAuth(self, authHeader):
        rD = {"errorCode": None, "errorMessage": None, "token": None, "errorFlag": False}
        parts = authHeader.split()
        if parts[0].lower() != "bearer":
            rD = {"errorCode": self.__tokenErrorCode, "errorMessage": "Authorization header must start with Bearer", "errorFlag": True}
        elif len(parts) == 1:
            rD = {"errorCode": self.__tokenErrorCode, "errorMessage": "API access token not found", "errorFlag": True}
        elif len(parts) > 2:
            rD = {"errorCode": self.__tokenErrorCode, "errorMesage": "Authorization header must be Bearer token", "errorFlag": True}
        if not rD["errorFlag"]:
            rD["token"] = parts[1]
        return rD

    def parseToken(self, token):
        """Return data payload for the input token along with any error diagnostics -"""
        return self.__getTokenData(token)

    def __parseToken(self, token, secretKey="secretvalue", algorithm="HS256"):
        return jwt.decode(token, secretKey, algorithms=algorithm)

    def __getTokenData(self, token):
        """Return token payload as a dictionary and processing diagnostics -"""
        errorCode = None
        errorMessage = None
        tokenData = {}
        #
        if token is None:
            errorMessage = "Missing token"
        try:
            tokenData = self.__parseToken(token, secretKey=self.__serviceKey)
        except jwt.DecodeError:
            errorMessage = "API access token is invalid"
            errorCode = self.__tokenErrorCode
        except jwt.ExpiredSignatureError:
            errorMessage = "API access token has expired"
            errorCode = self.__tokenErrorCode
        except:  # noqa: E722 pylint: disable=bare-except
            errorMessage = "API access token processing error"
            errorCode = self.__tokenErrorCode

        tokenData["errorCode"] = errorCode
        tokenData["errorMessage"] = errorMessage
        tokenData["errorFlag"] = errorMessage is not None

        return tokenData



##
# File:  ServiceSessionState
# Date:  2-July-2016
#
# Updated:
#         5-Aug-2016  jdw add support for application data
##
"""
Accessors to encapsulate common service session data management details --

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import logging

logger = logging.getLogger()


class ServiceSessionState(object):
    """Accessors to encapsulate common service session details --"""

    #

    def __init__(self):
        """"""
        #
        self.__D = {}
        self.__strKeyList = ["servicename", "serviceargs", "statustext", "errormessage", "warningmessage", "responseformat"]
        #
        self.__boolKeyList = ["errorflag", "warningflag"]
        self.clear()
        #
        #  Containers for variable application data and files details-
        self.__A = {}
        self.__uFL = []
        self.__dFL = []
        #
        self.clear()

    #

    def clear(self):
        self.__A = {}
        self.__uFL = []
        self.__dFL = []
        self.__D = {}
        for ky in self.__strKeyList:
            self.__D[ky] = ""
        for ky in self.__boolKeyList:
            self.__D[ky] = False

    def setAppDataDict(self, dictval, errFlag=False, format="json"):  # pylint: disable=redefined-builtin
        try:
            self.__D["responseformat"] = format
            self.__D["errorflag"] = errFlag
            for k, v in dictval.items():
                self.__A[k] = v
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    def setAppData(self, key, value):
        try:
            self.__A[key] = value
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    def getAppDataDict(self):
        return self.__A

    def getUploadList(self):
        return self.__uFL

    def getDownloadList(self):
        return self.__dFL

    def setDownload(self, fileName, filePath, contentType=None, md5Digest=None):
        try:
            self.__dFL.append((fileName, filePath, contentType, md5Digest))
            self.__D["responseformat"] = "files"
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    def setUpload(self, fileName, filePath, contentType=None, md5Digest=None):
        try:
            self.__uFL.append((fileName, filePath, contentType, md5Digest))
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    def assign(self, name, args=None, completionFlag=None):
        try:
            if name is not None:
                self.setServiceName(name)
            if args is not None:
                self.setServiceArgs(args)
            if completionFlag is not None:
                self.setServiceCompletionFlag(completionFlag)
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            logging.exception("FAILING")
            return False

    def setResponseFormat(self, format="json"):  # pylint: disable=redefined-builtin
        self.__D["responseformat"] = format

    def getResponseFormat(self):
        return self.__D["responseformat"]

    def setServiceName(self, name):
        self.__D["servicename"] = name

    def getServiceName(self):
        return self.__D["servicename"]

    def setServiceArgs(self, val):
        self.__D["serviceargs"] = val

    def getServiceArgs(self):
        return self.__D["serviceargs"]

    def setServiceStatusText(self, msg):
        self.__D["statustext"] = msg

    def getServiceStatusText(self):
        return self.__D["statustext"]

    def setServiceCompletionFlag(self, boolFlag):
        self.__D["errorflag"] = not boolFlag

    #
    def setServiceError(self, msg, errFlag=True, format="json"):  # pylint: disable=redefined-builtin
        self.__D["errorflag"] = errFlag
        self.__D["errormessage"] = msg
        self.__D["responseformat"] = format

    #

    def setServiceWarning(self, msg, warnFlag=True, format="json"):  # pylint: disable=redefined-builtin
        self.__D["warningflag"] = warnFlag
        self.__D["warningmessage"] = msg
        self.__D["responseformat"] = format

    def setServiceErrorFlag(self, boolFlag):
        self.__D["errorflag"] = boolFlag

    def getServiceErrorFlag(self):
        return self.__D["errorflag"]

    def setServiceWarningFlag(self, boolFlag):
        self.__D["warningflag"] = boolFlag

    def getServiceWarningFlag(self):
        return self.__D["warningflag"]

    def setServiceErrorMessage(self, msg):
        self.__D["errormessage"] = msg

    def getServiceErrorMessage(self):
        return self.__D["errormessage"]

    def setServiceWarningMessage(self, msg):
        self.__D["warningmessage"] = msg

    def getServiceWarningMessage(self):
        return self.__D["warningmessage"]


#



##
# File:    ServiceSessionFactory.py
# Date:    6-Jul-2016 J. Westbrook
#
# Updates:
##
"""
Utilities for service session directory management.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"


import hashlib
import time
import os.path
import shutil
import logging

logger = logging.getLogger()


class ServiceSessionFactory(object):
    """
    Utilities for service session directory management.

    """

    def __init__(self, topPath=None, serviceUserId=None):
        """
        Organization of the service session directory is --

        <topPath>/<Service_id>/<sha-hash>/<session_files>

        Parameters:
        :param string topPath: is the path to the directory containing the hash-id sub-directory.
        :param string serviceId: is the path to the directory containing the hash-id sub-directory.

        """
        self.__topSessionPath = topPath if topPath else "."
        self.__serviceUserId = serviceUserId if serviceUserId else "ANONYMOUS"
        self.__uid = None

    def __str__(self):
        return "Session top path: %s\nService user id: %s\nUnique identifier: %s\nSession path: %s\n" % (self.__topSessionPath, self.__serviceUserId, self.__uid, self.getPath())

    def __repr__(self):
        return self.__str__()

    def setId(self, uid):
        self.__uid = uid
        return True

    def getId(self):
        return self.__uid

    def assignId(self):
        self.__uid = hashlib.sha1(repr(time.time()).encode("utf-8")).hexdigest()
        return self.__uid

    def __getPath(self, relative=False):
        pth = None
        try:
            if relative:
                pth = os.path.join("/sessions", self.__serviceUserId, self.__uid)
            else:
                pth = os.path.join(self.__topSessionPath, "sessions", self.__serviceUserId, self.__uid)
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("FAILING")
            pth = None

        return pth

    def getPath(self):
        try:
            pth = self.__getPath()
            logger.debug("Session path %r", pth)
            if os.access(pth, os.F_OK):
                return pth
            else:
                return None
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("FAILING")

        return None

    def getRelativePath(self):
        return self.__getPath(relative=True)

    def getTopPath(self):
        return self.__topSessionPath

    def getSessionUserPath(self):
        return os.path.join(self.__topSessionPath, "sessions", self.__serviceUserId)

    def setTopPath(self, topSessionPath):
        self.__topSessionPath = topSessionPath
        return True

    def getTopSessionPath(self):
        return self.__topSessionPath

    def setTopSessionPath(self, topSessionPath):
        self.__topSessionPath = topSessionPath
        return True

    def setServiceUserId(self, serviceUserId):
        self.__serviceUserId = serviceUserId
        return True

    def makeSessionPath(self):
        """If the path to the current session directory does not exist
        create it and return the session path.
        """
        try:
            pth = self.__getPath()
            if not os.access(pth, os.F_OK):
                os.makedirs(pth)
            return pth
        except:  # noqa: E722 pylint: disable=bare-except
            return None

    def remakeSessionPath(self):
        try:
            pth = self.__getPath()
            if os.access(pth, os.F_OK):
                shutil.rmtree(pth, True)
            os.makedirs(pth)
            return pth
        except:  # noqa: E722 pylint: disable=bare-except
            return None



##
# File: ServiceLockFile.py
# Date: 21-Feb-2012  John Westbrook
#
# Update:
#      02-Aug-2016  jdw adapt for service application -  add logging -
#      14-Mar02017  jdw log aquire timeout
#
##
"""
Class implementing a cross-platform file locking strategy using an auxiliary lock file.

"""

import os
import time
import errno
import logging

logger = logging.getLogger()


class LockFileTimeoutException(Exception):
    pass


class ServiceLockFile(object):
    """A simple cross-platform file locking utilitiy class using an auxiliary
    lock file.

    This class provides internal methods supporting context-management
    (e.g. with statement support).  For example,

    with ServiceLockFile("target-file.db", timeoutSeconds=2) as lock:
        # - process/update the target-file.db.

    # - At the end of this 'with' clause the lock is automatically removed.

    """

    def __init__(self, filePath, timeoutSeconds=15, retrySeconds=0.2):
        """Prepare the file locker. Specify the file to lock and optionally
        the maximum timeoutSeconds and the retrySeconds between each attempt to lock.

        It is assumed that the locking file will be created within the
        path of the target file.
        """
        self.__isLocked = False
        self.__lockFilePath = os.path.join(filePath + ".lock")
        self.__filePath = filePath
        #
        self.__timeoutSeconds = timeoutSeconds
        self.__retrySeconds = retrySeconds
        self.__debug = True
        self.__fd = None
        #

    def acquire(self):
        """Create the lock if no lock file exists.  If a lockfile exists then
        repeat the test every retrySeconds.

        If the lock cannot be acquired within  'timeoutSeconds' then
        throw an exception.

        """
        timeBegin = time.time()
        while True:
            try:
                self.__fd = os.open(self.__lockFilePath, os.O_CREAT | os.O_EXCL | os.O_RDWR)
                if self.__debug:
                    logger.debug("Lock file created %s", self.__lockFilePath)
                break
            except OSError as myErr:
                if myErr.errno != errno.EEXIST:
                    # pass on some unanticipated problem ---
                    raise
                # handle timeout and retry -
                if (time.time() - timeBegin) >= self.__timeoutSeconds:
                    logger.debug("ServiceLockfile(acquire) Failed to acquire lock within timeout %r", self.__timeoutSeconds)
                    raise LockFileTimeoutException("ServiceLockFile(acquire) Internal timeout of %d (seconds) exceeded for %s" % (self.__timeoutSeconds, self.__filePath))
                if self.__debug:
                    logger.debug("Lock file retry for file %s", self.__lockFilePath)
                time.sleep(self.__retrySeconds)
        #
        self.__isLocked = True

    def release(self):
        """Remove any existing lock file."""
        if self.__isLocked:
            os.close(self.__fd)
            os.unlink(self.__lockFilePath)
            self.__isLocked = False
            if self.__debug:
                logger.debug("LockFile(release) removed lock file %s", self.__lockFilePath)

    def __enter__(self):
        """Internal method invoked at the beginning of a 'with' clause."""
        if not self.__isLocked:
            self.acquire()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Internal method invoked at the end of a 'with' clause."""
        if self.__isLocked:
            self.release()

    def __del__(self):
        """Internal method to cleanup any lingering lock file."""
        self.release()


if __name__ == "__main__":
    pass



##
# File:    ServiceResponse.py
# Date:    9-July-2016  J. Westbrook
#
# Updated:
#
##
"""
Containers and accessors for managing responses to web service requests.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"


import os
import gzip
import mimetypes
import logging
from webob import Response


try:
    import json
except ImportError:  # pragma: no cover
    import simplejson as json

try:
    # Python 2
    text_type = unicode
except NameError:
    # Python 3
    text_type = str

logger = logging.getLogger()


class ServiceResponse(object):
    def __init__(self, returnFormat="json", injectStatus=True):
        """
        Manage content items to be transfered as part of the application response.

        Set injectStatus=True to add statustext and errorflag to the data payload
          if these are not set.

        """
        self.__injectStatus = injectStatus
        self._cD = self.__setup(returnFormat)

    def __setup(self, returnFormat):
        """Default response content is set here."""
        cD = {}
        cD["returnformat"] = returnFormat
        cD["htmllinkcontent"] = ""
        cD["htmlcontent"] = ""
        cD["textcontent"] = ""
        cD["location"] = ""
        cD["contentmimetype"] = None
        cD["encodingtype"] = None
        cD["disposition"] = None
        #
        # Data payload container for file based data content
        cD["datafilecontent"] = None
        cD["datafilename"] = None
        cD["datafilechecksum"] = None
        #
        #  Data payload for all json objects -
        cD["datacontent"] = {}

        #  Status items
        cD["errorflag"] = False
        cD["statustext"] = ""
        cD["statuscode"] = 200
        #
        #
        return cD

    def isError(self):
        return self._cD["errorflag"]

    def setError(self, statusCode=200, msg=""):
        self._cD["errorflag"] = True
        self._cD["statustext"] = msg
        self._cD["statuscode"] = statusCode

    #
    def setData(self, dataObj=None):
        self._cD["datacontent"] = dataObj

    def getData(self):
        return self._cD["datacontent"]

    #
    def setHtmlList(self, htmlList=None):
        if htmlList is None:
            htmlList = []
        self._cD["htmlcontent"] = "\n".join(htmlList)

    def appendHtmlList(self, htmlList=None):
        if htmlList is None:
            htmlList = []
        if len(self._cD["htmlcontent"]) > 0:
            self._cD["htmlcontent"] = "%s\n%s" % (self._cD["htmlcontent"], "\n".join(htmlList))
        else:
            self._cD["htmlcontent"] = "\n".join(htmlList)

    def setHtmlText(self, htmlText=""):
        self._cD["htmlcontent"] = htmlText

    def setHtmlTextFromTemplate(self, templateFilePath, webIncludePath, parameterDict=None, insertContext=False):
        pD = parameterDict if parameterDict is not None else {}
        self._cD["htmlcontent"] = self.__processTemplate(templateFilePath=templateFilePath, webIncludePath=webIncludePath, parameterDict=pD, insertContext=insertContext)

    def setHtmlLinkText(self, htmlText=""):
        self._cD["htmllinkcontent"] = htmlText

    def setText(self, text=""):
        self._cD["textcontent"] = text

    def setLocation(self, url=""):
        self._cD["location"] = url

    def setTextFile(self, filePath):
        try:
            if os.path.exists(filePath):
                with open(filePath, "r") as fin:
                    self._cD["textcontent"] = fin.read()
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("+setTextFile() File read failed %s\n", filePath)

    def setHtmlContentPath(self, aPath):
        self._cD["htmlcontentpath"] = aPath

    def getMimetypeAndEncoding(self, filename):
        ftype, encoding = mimetypes.guess_type(filename)
        # We'll ignore encoding, even though we shouldn't really
        if ftype is None:
            if filename.find(".cif.V") > 0:
                ret = ("text/plain", None)
            else:
                ret = ("application/octet-stream", None)
        else:
            ret = (ftype, encoding)
        return ret

    def setBinaryFile(self, filePath, attachmentFlag=False, serveCompressed=True, md5Digest=None):
        try:
            if os.path.exists(filePath):
                _dir, fn = os.path.split(filePath)
                if not serveCompressed and fn.endswith(".gz"):
                    with gzip.open(filePath, "rb") as fin:
                        self._cD["datafilecontent"] = fin.read()
                    self._cD["datafileName"] = fn[:-3]
                    contentType, encodingType = self.getMimetypeAndEncoding(filePath[:-3])
                else:
                    with open(filePath, "rb") as fin:
                        self._cD["datafilecontent"] = fin.read()
                    self._cD["datafileName"] = fn
                    contentType, encodingType = self.getMimetypeAndEncoding(filePath)
                #
                self._cD["contentmimetype"] = contentType
                self._cD["encodingtype"] = encodingType
                if attachmentFlag:
                    self._cD["disposition"] = "attachment"
                else:
                    self._cD["disposition"] = "inline"
                    #
                    # strip compression file extension if disposition=inline.
                    if fn.endswith(".gz"):
                        self._cD["datafileName"] = fn[:-3]
                if md5Digest:
                    self._cD["datafilechecksum"] = md5Digest
                logger.debug("Serving %s as %s encoding %s att flag %r checksum %r\n", filePath, contentType, encodingType, attachmentFlag, md5Digest)
                return True
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("ResponseContent.setBinaryFile() File read failed %s\n", filePath)
        return False

    def wrapFileAsJsonp(self, filePath, callBack=None):
        try:
            if os.path.exists(filePath):
                _dir, fn = os.path.split(filePath)
                (_rn, ext) = os.path.splitext(fn)
                #
                dd = {}
                with open(filePath, "r") as fin:
                    dd["data"] = fin.read()
                if ext.lower() != ".json":
                    self._cD["datafilecontent"] = callBack + "(" + json.dumps(dd) + ");"
                else:
                    self._cD["datafilecontent"] = callBack + "(" + dd["data"] + ");"
                #
                self._cD["datafileName"] = fn
                contentType = "application/x-javascript"
                encodingType = None
                #
                self._cD["contentmimetype"] = contentType
                self._cD["encodingtype"] = encodingType
                self._cD["disposition"] = "inline"
                #

                logger.debug("Serving %s as %s\n", filePath, self._cD["datafilecontent"])
        except Exception as e:
            logging.exception("File read failed %s err %r", filePath, str(e))

    def dump(self, maxLength=130):
        retL = []
        retL.append("Service response object:")
        for k, v in self._cD.items():
            if v is None:
                continue
            elif isinstance(v, dict):
                retL.append("  - key = %-35s - dict : %s" % (k, v.items()))
            elif v is not None and len(str(v).strip()) > 0:
                retL.append("  - key = %-35s - value(1-%d): %s" % (k, maxLength, str(v)[:maxLength]))
        return "\n   ".join(retL)

    def setReturnFormat(self, format):  # pylint: disable=redefined-builtin
        if format in ["html", "text", "json", "jsonText", "jsonData", "location", "binary", "jsonp"]:
            self._cD["returnformat"] = format
            return True
        else:
            return False

    def __injectMessage(self, tag, msg):
        try:
            if tag not in self._cD["datacontent"]:
                self._cD["datacontent"][tag] = msg
        except:  # noqa: E722 pylint: disable=bare-except
            pass
        return False

    def getResponse(self):
        rspD = self.__getD()
        #
        #  Build the WebOb response -
        #
        myResponse = Response()
        myResponse.status = rspD["STATUS_CODE"]
        myResponse.content_type = rspD["CONTENT_TYPE"]

        if isinstance(rspD["RETURN_STRING"], text_type):
            myResponse.text = rspD["RETURN_STRING"]
        else:
            myResponse.body = rspD["RETURN_STRING"]

        if "ENCODING" in rspD:
            myResponse.content_encoding = rspD["ENCODING"]
        if "DISPOSITION" in rspD:
            myResponse.content_disposition = rspD["DISPOSITION"]
        if "CHECKSUM_MD5" in rspD:
            myResponse.headers.add("CHECKSUM_MD5", rspD["CHECKSUM_MD5"])
        #
        return myResponse

    def __getD(self):
        """Return an internal dictionary as precursor to preparing for web server response object"""
        rD = {}
        returnFormat = self._cD["returnformat"]

        #
        # Handle error cases --
        #
        if self._cD["errorflag"]:
            if returnFormat in ["json", "jsonText", "jsonData", "jsonp"]:
                if self.__injectStatus:
                    self.__injectMessage("statustext", self._cD["statustext"])
                    self.__injectMessage("errorflag", self._cD["errorflag"])

                rD = self.__initJsonResponse(self._cD["datacontent"])
            elif returnFormat in ["html", "location"]:
                rD = self.__initHtmlResponse(self._cD["statustext"])
            elif returnFormat in ["text", "binary"]:
                rD = self.__initTextResponse(self._cD["statustext"])
            else:
                rD = self.__initTextResponse(self._cD["statustext"])
        else:
            #
            if returnFormat == "html":
                rD = self.__initHtmlResponse(self._cD["htmlcontent"])
            #
            elif returnFormat == "text":
                rD = self.__initTextResponse(self._cD["textcontent"])
            #
            elif returnFormat == "location":
                rD = self.__initLocationResponse(self._cD["location"])
            #
            elif returnFormat == "jsonText":
                rD = self.__initJsonResponseInTextArea(self._cD["datacontent"])
            #
            elif returnFormat == "json":
                rD = self.__initJsonResponse(self._cD["datacontent"])
            #
            elif returnFormat == "jsonData":
                rD = self.__initJsonResponse(self._cD["datacontent"])
            #
            elif returnFormat == "binary":
                rD = self.__initBinaryResponse(self._cD)
                if self._cD["datafilechecksum"]:
                    rD["CHECKSUM_MD5"] = self._cD["datafilechecksum"]
            #
            elif returnFormat == "jsonp":
                rD = self.__initJsonpResponse(self._cD)
            else:
                pass
        #
        rD["STATUS_CODE"] = self._cD["statuscode"]
        return rD

    def __initLocationResponse(self, url):
        rspDict = {}
        rspDict["CONTENT_TYPE"] = "location"
        rspDict["RETURN_STRING"] = url
        return rspDict

    def __initBinaryResponse(self, myD=None):
        if myD is None:
            myD = {}
        rspDict = {}
        rspDict["CONTENT_TYPE"] = myD["contentmimetype"]
        rspDict["RETURN_STRING"] = myD["datafilecontent"]
        try:  # noqa: E722 pylint: disable=bare-except
            rspDict["ENCODING"] = myD["encodingtype"]
            if myD["disposition"] is not None:
                rspDict["DISPOSITION"] = "%s; filename=%s" % (myD["disposition"], myD["datafileName"])
        except:  # noqa: E722 pylint: disable=bare-except
            pass
        return rspDict

    def __initJsonResponse(self, myD=None):
        if myD is None:
            myD = {}
        rspDict = {}
        rspDict["CONTENT_TYPE"] = "application/json"
        rspDict["RETURN_STRING"] = json.dumps(myD)
        return rspDict

    def __initJsonpResponse(self, myD=None):
        if myD is None:
            myD = {}
        rspDict = {}
        rspDict["CONTENT_TYPE"] = myD["contentmimetype"]
        rspDict["RETURN_STRING"] = myD["datafilecontent"]
        return rspDict

    def __initJsonResponseInTextArea(self, myD=None):
        if myD is None:
            myD = {}
        rspDict = {}
        rspDict["CONTENT_TYPE"] = "text/html"
        rspDict["RETURN_STRING"] = "<textarea>" + json.dumps(myD) + "</textarea>"
        return rspDict

    def __initHtmlResponse(self, myHtml=""):
        rspDict = {}
        rspDict["CONTENT_TYPE"] = "text/html"
        rspDict["RETURN_STRING"] = myHtml
        return rspDict

    def __initTextResponse(self, myText=""):
        rspDict = {}
        rspDict["CONTENT_TYPE"] = "text/plain"
        rspDict["RETURN_STRING"] = myText
        return rspDict

    def __processTemplate(self, templateFilePath="./alignment_template.html", webIncludePath=".", parameterDict=None, insertContext=False):
        """Read the input HTML template data file and perform the key/value substitutions in the
        input parameter dictionary.

        if insertContext is set then paramDict is injected as a json object if <!--insert application_context=""-->

        Template HTML file path -  (e.g. /../../htdocs/<appName>/template.html)
        webTopPath = file system path for web includes files  (eg. /../../htdocs) which will
                     be prepended to embedded include path in the HTML template document
        """

        if parameterDict is None:
            parameterDict = {}
        try:
            ifh = open(templateFilePath, "r")
            sL = []
            for line in ifh.readlines():
                if str(line).strip().startswith("<!--#include") or (insertContext and str(line).strip().startswith("<!--#insert")):
                    fields = str(line).split('"')
                    tpth = os.path.join(webIncludePath, fields[1][1:])
                    try:
                        tfh = open(tpth, "r")
                        sL.append(tfh.read())
                        tfh.close()
                    except Exception as e:
                        logger.debug("failed to include %s fields=%r err=%r\n", tpth, fields, str(e))
                else:
                    sL.append(line)
            ifh.close()
            sIn = "".join(sL)
            return sIn % parameterDict
        except:  # noqa: E722 pylint: disable=bare-except
            logger.exception("Failed for %s\n", templateFilePath)

        return ""
